-- FSL Fast I/O controller
--
-- Copyright (C) 2004-2005, Claudio Lanconelli
--		All rights reserved.
--
-- http://www.lancos.com
--

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
--library UNISIM;
--use UNISIM.VComponents.all;

entity fsl_ioctrl is
	GENERIC (
		FSL_DATA_BITS : INTEGER := 16;
		OUTPUT_BITS : INTEGER := 8;
		INPUT_BITS : INTEGER := 8;
		START_BITS : INTEGER := 0;
		START_CMD_RISING : INTEGER := 0
	);
    PORT (
		clk : in std_logic;
		reset : in std_logic;

		-- I/O
		inputs : in std_logic_vector(INPUT_BITS-1 downto 0);
		outputs : out std_logic_vector(OUTPUT_BITS-1 downto 0);

		-- CMD CTRL
		cmd_end : in std_logic_vector(START_BITS downto 0);

		--FSL SLAVE lines
		FSL_S_Clk : out std_logic;
		FSL_S_Data : in std_logic_vector(0 to FSL_DATA_BITS-1);
		FSL_S_Control : in std_logic;
		FSL_S_Read : out std_logic;
		FSL_S_Exists : in std_logic;

		--FSL MASTER lines
		FSL_M_Clk : out std_logic;
		FSL_M_Write : out std_logic;
		FSL_M_Data : out std_logic_vector(0 to FSL_DATA_BITS-1);
		FSL_M_Control : out std_logic;
		FSL_M_Full : in std_logic
	);
end fsl_ioctrl;

architecture Behavioral of fsl_ioctrl is

	CONSTANT IOADDR_BITS : INTEGER := 2;

	CONSTANT C_READ : std_logic := '1';
	CONSTANT C_WRITE : std_logic := '0';

	signal sig_clk : std_logic;
	signal sig_reset : std_logic;

	signal sig_cmd_end : std_logic_vector(START_BITS downto 0);
	signal sampled_cmd_end : std_logic_vector(START_BITS downto 0);

	signal sig_outputs : std_logic_vector(OUTPUT_BITS-1 downto 0);

	signal sig_ctrlin : std_logic;
	signal sig_datain : std_logic_vector(FSL_DATA_BITS-1 downto 0);
	signal sig_ioaddr : std_logic_vector(IOADDR_BITS-1 downto 0);
	signal sig_rnw : std_logic;
	signal sig_dataclk : std_logic;

	signal sig_fsli_data : std_logic_vector(FSL_DATA_BITS-1 downto 0);
	signal sig_read_req : std_logic;
begin
	sig_clk <= clk;
	sig_reset <= reset;

	FSL_S_Clk <= sig_clk;
	FSL_M_Clk <= sig_clk;

	gen_data_mfsl:
	for bit_idx in 0 to (FSL_DATA_BITS-1) generate
	begin
		FSL_M_Data(FSL_DATA_BITS-1 - bit_idx) <= sig_fsli_data(bit_idx);
	end generate;
--	FSL_M_Data <= sig_fsli_data;
	FSL_M_Control <= '0';	--non usato

	outputs <= sig_outputs;

	sig_cmd_end <= cmd_end;

	-- Legge tutti i dati dalla FSL come si presentano
	FSL_S_Read <= FSL_S_Exists;
	sig_dataclk <= FSL_S_Exists;

	gen_data_sfsl:
	for bit_idx in 0 to (FSL_DATA_BITS-1) generate
	begin
		sig_datain(bit_idx) <= FSL_S_Data(FSL_DATA_BITS-1 - bit_idx);
	end generate;
--	sig_datain <= FSL_S_Data;
	sig_ctrlin <= FSL_S_Control;

	sig_ioaddr <= sig_datain(FSL_DATA_BITS-1 downto FSL_DATA_BITS-IOADDR_BITS);
	sig_rnw <= sig_ctrlin;

	FSL_M_Write <= sig_read_req and not FSL_M_Full;

	gen_intflags: process (sig_clk)
	begin
	if rising_edge(sig_clk) then
		if sig_reset = '1' then
			sig_outputs <= (others => '0');
			sig_read_req <= '0';
			sig_fsli_data <= (others => '0');
			sampled_cmd_end <= (others => '0');
		else
			if START_BITS > 0 then
				sampled_cmd_end <= sig_cmd_end;

				for sig_index in 0 to START_BITS-1 loop
					if START_CMD_RISING = 0 then
						if sig_cmd_end(sig_index) = '0' and sampled_cmd_end(sig_index) = '1' then	--falling edge
							sig_outputs(sig_index) <= '0';
						end if;
					else
						if sig_cmd_end(sig_index) = '1' and sampled_cmd_end(sig_index) = '0' then	--rising edge
							sig_outputs(sig_index) <= '0';
						end if;
					end if;
				end loop;
			end if;

			sig_read_req <= '0';

			if sig_dataclk = '1' then

				if sig_rnw = C_WRITE then
					case sig_ioaddr(IOADDR_BITS-1 downto 0) is	--WRITE
						when "00" =>		-- CLR
							if OUTPUT_BITS > 0 then
								sig_outputs <= sig_outputs and not sig_datain(OUTPUT_BITS-1 downto 0);
							end if;
						when "01" =>		-- SET
							if OUTPUT_BITS > 0 then
								sig_outputs <= sig_outputs or sig_datain(OUTPUT_BITS-1 downto 0);
							end if;
						when "10" =>		-- ASSIGN
							if OUTPUT_BITS > 0 then
								sig_outputs <= sig_datain(OUTPUT_BITS-1 downto 0);
							end if;
						when "11" =>		-- TOGGLE
							if OUTPUT_BITS > 0 then
								sig_outputs <= sig_outputs xor sig_datain(OUTPUT_BITS-1 downto 0);
							end if;
						when others =>
							null;
					end case;
				else
					sig_read_req <= '1';

					sig_fsli_data <= (others => '0');
					sig_fsli_data(FSL_DATA_BITS-1 downto FSL_DATA_BITS-IOADDR_BITS) <= sig_ioaddr;

					case sig_ioaddr(IOADDR_BITS-1 downto 0) is	--READ
						when "00" =>	--Read Back outputs
							if OUTPUT_BITS > 0 then
								sig_fsli_data(OUTPUT_BITS-1 downto 0) <= sig_outputs;
							end if;
						when "01" =>	--Read inputs
							if INPUT_BITS > 0 then
								sig_fsli_data(INPUT_BITS-1 downto 0) <= inputs;
							end if;
						when "10" =>	--Read inputs + write outputs
							if OUTPUT_BITS > 0 then
								sig_fsli_data(INPUT_BITS-1 downto 0) <= inputs;
								sig_outputs <= sig_datain(OUTPUT_BITS-1 downto 0);
							end if;
						when "11" =>	--Read inputs + toggle outputs
							if INPUT_BITS > 0 then
								sig_fsli_data(INPUT_BITS-1 downto 0) <= inputs;
								sig_outputs <= sig_outputs xor sig_datain(OUTPUT_BITS-1 downto 0);
							end if;
						when others =>
							null;
					end case;
				end if;
			end if;
		end if;
	end if;
	end process;

end Behavioral;
